"""Top-level package for CARSpy."""
from .line_strength import LineStrength
from .cars_synth import CarsSpectrum
from .cars_fit import CarsFit
from .utils import *

__author__ = """Zhiyao Yin"""
__email__ = 'zhiyao.yin@dlr.de'
__version__ = '0.6.0'



"""Additional functions used within the CARSpy program."""
from functools import wraps
import pickle
import numpy as np


try:
    import matplotlib.pyplot as plt
    from scipy.signal import find_peaks
    _HAS_SCIPY_PYPLOT = True
except Exception:
    _HAS_SCIPY_PYPLOT = False


def _ensureModules(function):
    if _HAS_SCIPY_PYPLOT:
        @wraps(function)
        def wrapper(*args, **kwargs):
            return function(*args, **kwargs)
        wrapper.__doc__ = function.__doc__
        return wrapper

    def no_op(*args, **kwargs):
        _message = ("Scipy and Matplotlib modules are required to carry out "
                    "peak finding")
        raise Exception(_message)
    no_op.__doc__ = function.__doc__
    return no_op


try:
    import cantera as ct
    _HAS_CANTERA = True
except Exception:
    _HAS_CANTERA = False


def _ensureCantera(function):
    if _HAS_CANTERA:
        @wraps(function)
        def wrapper(*args, **kwargs):
            return function(*args, **kwargs)
        wrapper.__doc__ = function.__doc__
        return wrapper

    def no_op(*args, **kwargs):
        _message = ("Cantera module is required for calculating equilibrium "
                    "composition. Please install cantera first or specify "
                    "custom eq_func in signal_as() instead.")
        raise Exception(_message)
    no_op.__doc__ = function.__doc__
    return no_op


@_ensureCantera
def eq_comp(temperature, pressure, init_comp, valid_from=1200.):
    """
    Calculate equilibrium composition at given temperature and pressure.

    .. attention::
        This function is only intended as a "dummy" template for setting up
        custom equilibrium solvers with ``cantera``. Please be aware of the
        applicabilities and uncertainties of various kinetic mechanisms.

    Parameters
    ----------
    temperature : float
        Temperature in [K].
    pressure : 1, float, optional
        Pressure in bars.
    init_comp : dict
        Initial gas mole fractions in a dictionary.
    valid_from : float
        Temperature lower boundary for when the mechanism is valid, default is
        1200 [K]. This value is by no means valid for all cases.
    """
    products = init_comp.copy()
    if temperature > valid_from:
        gas = ct.Solution('gri30.xml')
        gas.TPX = temperature, pressure*1e5, init_comp
        gas.equilibrate('TP')
        products = gas.mole_fraction_dict()

    # remove small values
    products = {key: products[key] for key in products if products[key] > 1e-5}

    return products


def downsample(w, w_fine, spec_fine, mode='local_mean'):
    """Downsample a fine spectrum according to specified coarse spectral axis.

    Parameters
    ----------
    w : sorted 1-D array of floats
        Coarse spectral axis (must be sorted and evenly spaced).
    w_fine : sorted 1-D array of floats
        Fine spectral axis (must be sorted and evenly spaced).
    spec_fine : 1-D array of floats
        Spectrum with fine resolution, must be of the same size as w_fine.
    mode : str, optional
        Two modes to choose from: 'local-mean' or 'interp', by default
        'local_mean'.

    Returns
    -------
    1-D array of floats
        Downsampled spectrum of the same size as w.
    """
    downsampled = []
    if mode == 'interp':
        downsampled = np.interp(w, w_fine, spec_fine)
    elif mode == 'local_mean':
        # downsample scale
        hw = int((w[1] - w[0])/(w_fine[1] - w_fine[0])/2)
        # search for closest indices
        w_fine = np.array(w_fine)
        idx = np.searchsorted(w_fine, w)
        idx[w_fine[idx] - w > np.diff(w_fine).mean()*0.5] -= 1
        # take local average based on the downsample scale
        downsampled = np.mean(
            [spec_fine[idx-_step] for _step in range(-hw, hw+1)], axis=0)

    return downsampled


def comp_normalize(comp_dict, target=1.0):
    """Normalize gas composition saved in a dictionary.

    Parameters
    ----------
    comp_dict : dict
        Gas composition in the measurement volume stored in a dictionary.
    target : float, optional
        Normalization factor, by default 1.0.

    Returns
    -------
    dict
        Normalized gas composition stored in a dictionary.
    """
    raw = sum(comp_dict.values())
    factor = target/raw
    return {key: value*factor for key, value in comp_dict.items()}


@_ensureModules
def loc_lines(spec, height, w_peaks, peak_indices=None, inspect=True):
    """[summary]

    Parameters
    ----------
    spec : 1d array of floats
        Spectrum containing distinct peaks.
    height : float
        A threshold for the peaks. Chose a value between 0 and 1.
    peak_indices : list of integer
        List of indices of peaks to use for calculating conversion factor.
    w_peaks : 1d array of floats
        List of targeted peaks in Angstrom.
    inpsect : bool, optional
        If true, the peaks will be displayed on top of the original spectrum.
    """
    peaks, _ = find_peaks(spec/spec.max(), height=height)
    if inspect:  # Inspect the peak finding results
        plt.plot(spec, 'k')
        plt.plot(peaks, spec[peaks], 'ro')
        plt.title('Close this window to continue the program')
        plt.show()

    if len(peaks) != len(w_peaks):
        raise ValueError('Adjust height to find the correct number \
                            of peaks (%d)' % (len(w_peaks),))

    if peak_indices is None:
        peak_indices = range(len(w_peaks))
    elif len(peaks) < len(peak_indices):
        raise ValueError('Number of indices exceed the peaks found.')

    conv_factor = (np.diff(w_peaks[peak_indices])
                   / np.diff(peaks[peak_indices]))
    offset = (w_peaks - peaks*conv_factor.mean())[peak_indices]
    # calculate spectral domain in nm
    w = (np.arange(len(spec))*conv_factor.mean() + offset.mean())/10

    return w, conv_factor, offset, peaks


def pkl_dump(path_write, data):
    """Dump data into a pickle file.

    Parameters
    ----------
    path_write : path
        Absolute path to the pickle file to be created.
    data: python object
    """
    with open(path_write, 'wb') as pf:
        pickle.dump(data, pf)


def pkl_load(path_load):
    """Load a data into from pickle file.

    Parameters
    ----------
    dir_save : path
        Absolute path to the pickle file to be loaded.
    """
    with open(path_load, 'rb') as pf:
        data = pickle.load(pf)

    return data



import os
import json


def mol_const(species='N2'):
    """Load molecular constants

    Parameters
    ----------
    species : str, optional
        Target molecular species, by default 'N2'. Currently only 'N2' is
        available
    """
    fpath = os.path.join(os.path.dirname(__file__), "_MOL_CONST.json")
    with open(fpath, 'r') as f:
        _mol_const = json.load(f)[species]

    # extract only  the necessary ones
    _keys = ['we', 'wx', 'wy', 'wz', 'Be', 'De', 'alpha_e', 'beta_e',
             'gamma_e', 'H0', 'He', 'mu', 'MW', 'Const_Raman', 'G/A']
    _mol_const = {_key: _mol_const[_key]['value'] for _key in _keys}

    return _mol_const


def univ_const(dict_key='Const_N'):
    """Load universal constants

    Parameters
    ----------
    dict_key : str, optional
        Specify the item to load from default _constants/_UNIV_CONST.json, by
        default 'Const_N'
    """
    fpath = os.path.join(os.path.dirname(__file__), "_UNIV_CONST.json")
    with open(fpath, 'r') as f:
        _mol_const = json.load(f)

    return _mol_const[dict_key]['value']


def chi_const(use_set='SET 1'):
    """Load third order nonresonant susceptibilities

    Parameters
    ----------
    use_set : str, optional
        Choose from the available set of susceptibilities, by default 'SET 1':
        - SET 1: based on CARSFT
        - SET 2: based on NRC
        - SET 3: based on Eckbreth
    """
    fpath = os.path.join(os.path.dirname(__file__), "_CHI_NRS.json")
    with open(fpath, 'r') as f:
        _chi_const = json.load(f)

    return _chi_const[use_set]


def get_const(const_type='mol', show=False):
    """Get default constants stored under _constants/

    Parameters
    ----------
    const_type : str, optional
        Type of constants to print. Choose from 'mol', 'univ' and 'chi',
        by default 'mol'
    """
    fname = []
    if const_type == 'mol':
        fname = "_MOL_CONST.json"
    elif const_type == 'univ':
        fname = "_UNIV_CONST.json"
    elif const_type == 'chi':
        fname = "_CHI_NRS.json"

    fpath = os.path.join(os.path.dirname(__file__), fname)
    with open(fpath, 'r') as f:
        _const = json.load(f)

    if show:
        print(json.dumps(_const, indent=4))
    else:
        return _const



#!/usr/bin/env python

"""The setup script."""
from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['numpy']

setup_requirements = []

test_requirements = ['numpy']

setup(
    author="Zhiyao Yin",
    author_email='zhiyao.yin@dlr.de',
    python_requires='>=3.5',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description=("Synthesizing and fitting coherent anti-Stokes Raman spectra "
                 "(CARS) in Python"),
    install_requires=requirements,
    license="BSD license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='carspy',
    name='carspy',
    packages=find_packages(include=['carspy', 'carspy.*']),
    package_data={'carspy': ['_constants/*.json']},
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/chuckedfromspace/carspy',
    version='0.6.0',
    zip_safe=False,
)



"""Functions used in the convolution of CARS spectrum.

- Laser lineshape
- Impulse spectral response function (ISRF) for the spectrometer slit
"""
import numpy as np


def gaussian_line(w, w0, sigma):
    """Generate a normalized Gaussian lineshape (integral equals to 1).

    Parameters
    ----------
    w : sorted 1d array of floats
        Spectral positions in wavenumber cm^(-1).
    w0 : float
        Center of the Gaussian lineshape in wavenumber cm^(-1).
    sigma : float
        FWHM of the Gaussian lineshape wavenumber cm^(-1).

    Returns
    -------
    1d array of floats
        Intensities of the normalized Gaussian lineshape over w.
    """
    _lineshape = 2/sigma*(np.log(2)/np.pi)**0.5*np.exp(
        -4*np.log(2)*((w-w0)/sigma)**2)

    return _lineshape


def lorentz_line(w, w0, sigma):
    """Generate a normalized Lorentzian lineshape (integral equals to 1).

    Parameters
    ----------
    w : sorted 1d array of floats
        Spectral positions in wavenumber cm^(-1).
    w0 : float
        Center of the Lorentzian lineshape in wavenumber cm^(-1).
    sigma : float
        FWHM of the Lorentzian lineshape wavenumber cm^(-1).

    Returns
    -------
    1d array of floats
        Intensities of the normalized Lorentzian lineshape over w.
    """
    _lineshape = 1/np.pi*(sigma/2)/((w-w0)**2+sigma**2/4)

    return _lineshape


def voigt_line(w, w0, sigma_V, sigma_L):
    """Generate an approximated Voigt lineshape following :cite:`Whiting:68`.

    Parameters
    ----------
    w : 1d array of floats
        Spectral positions in wavenumber cm^(-1).
    w0 : float
        Center of the Lorentzian lineshape in wavenumber cm^(-1).
    sigma_V : float
        FWHM of the Voigt lineshape wavenumber cm^(-1).
    sigma_L : float
        FWHM of the Lorentzian lineshape wavenumber cm^(-1).

    Returns
    -------
    1d array
        Intensities of the Voigt lineshape over w.
    """
    # Preparations
    _ratio = sigma_L/sigma_V
    I_g = 1/(sigma_V*(1.065 + 0.447*_ratio + 0.058*_ratio**2))
    _w = abs(w-w0)/sigma_V
    # Building up the function
    _term_1 = I_g*(1-_ratio)*np.exp(-2.772*_w**2) + _ratio/(1 + 4*_w**2)
    _term_2 = 0.016*(1-_ratio)*_ratio*(np.exp(-0.4*_w**2.25)
                                       - 10/(10 + _w**2.25))

    return _term_1 + _term_2


def asym_Gaussian(w, w0, sigma, k, a_sigma, a_k, offset):
    """Asymmetric super-Gaussian following :cite:`Beirle:17`.

    Parameters
    ----------
    w : sorted 1d array of floats
        Spectral positions in wavenumber cm^(-1).
    w0 : float
        Center of the asymmetric Gaussian function in wavenumber cm^(-1).
    sigma : float
        FWHM of the Gaussian function in wavenumber cm^(-1).
    k : float
        Controls the skewing of the asymmetry.
    a_sigma, a_k : float
        Tuning factors for sigma and k.
    offset : float
        Background offset (from experimental spectrum).

    Returns
    -------
    1d array of floats
        Intensities of the peak-normalized asymmetric super-Gaussian over w.
    """
    response_low = np.exp(-abs((w[w <= w0]-w0)/(sigma-a_sigma))**(k-a_k))
    response_high = np.exp(-abs((w[w > w0]-w0)/(sigma+a_sigma))**(k+a_k))
    response = np.append(response_low, response_high) + offset

    return response/response.max()


def asym_Voigt(w, w0, sigma, k, a_sigma, a_k, sigma_L_l, sigma_L_h, offset):
    """Asymmetric super-Voigt.

    .. note::

        This is based on the super-Gaussian from :cite:`Beirle:17`, with
        additional convolution with two Lorentzian profiles to better capture
        slow-decaying wings in some experimental slit function

    Parameters
    ----------
    w : sorted 1d array of floats
        Spectral positions in wavenumber cm^(-1).
    w0 : float
        Center of the asymmetric Gaussian function in wavenumber cm^(-1).
    sigma : float
        FWHM of the Gaussian function in wavenumber cm^(-1).
    k : float
        Controls the skewing of the asymmetry.
    a_sigma, a_k : float
        Tuning factors for sigma and k.
    sigma_L_l : float
        FWHM of the Lorentzian function in wavenumber cm^(-1) for the
        lower half.
    sigma_L_h : float
        FWHM of the Lorentzian function in wavenumber cm^(-1) for the
        higher half.
    offset : float
        Background offset.

    Returns
    -------
    1d array of floats
        Intensities of the peak-normalized asymmetric super-Gaussian over w.
    """
    response_low = np.exp(-abs((w-w0)/(sigma-a_sigma))**(k-a_k))
    response_high = np.exp(-abs((w-w0)/(sigma+a_sigma))**(k+a_k))
    response_low = np.convolve(response_low, lorentz_line(w, w0, sigma_L_l),
                               'same')
    response_high = np.convolve(response_high, lorentz_line(w, w0, sigma_L_h),
                                'same')

    response = np.append(response_low[np.where(w <= w0)],
                         response_high[np.where(w > w0)]) + offset

    return response/response.max()


def asym_Voigt_deprecated(w, w0, sigma_V_l, sigma_V_h, sigma_L_l, sigma_L_h,
                          offset):
    """Asymmetric Voigt profile following NRC.

    .. admonition:: Deprecated
       :class: attention

       This profile cannot capture certain slit functions with broadened
       Gaussian profile.

    Parameters
    ----------
    w : sorted 1d array of floats
        Spectral positions in wavenumber cm^(-1).
    w0 : float
        Center of the asymmetric Gaussian function in wavenumber cm^(-1).
    sigma_V_l : float
        FWHM of the Voigt function in wavenumber cm^(-1) for the lower half.
    sigma_V_h : float
        FWHM of the Voigt function in wavenumber cm^(-1) for the higher half.
    sigma_L_l : float
        FWHM of the Lorentzian function in wavenumber cm^(-1) for the
        lower half.
    sigma_L_h : float
        FWHM of the Lorentzian function in wavenumber cm^(-1) for the
        higher half.
    offset : float
        Background offset.

    Returns
    -------
    1d array of floats
        Intensities of the peak-normalized asymmetric super-Gaussian over w.
    """
    response_low = voigt_line(w[w <= w0], w0, sigma_V_l, sigma_L_l)
    response_high = voigt_line(w[w > w0], w0, sigma_V_h, sigma_L_h)
    response = (np.append(response_low/response_low.max(),
                          response_high/response_high.max()) +
                offset)

    return response/response.max()


def slit_ISRF(w, w0, param_1, param_2, param_3, param_4, param_5, param_6,
              offset, mode='sGaussian'):
    """Impulse spectral response function (ISRF) as the slit function.

    Parameters
    ----------
    w : sorted 1d array of floats
        Spectral positions in wavenumber cm^(-1).
    w0 : float
        Center of the asymmetric Gaussian function in wavenumber cm^(-1).
    param_1, param_2, param_3, param_4 : float
        Parameters needed for the asymmetric ISRF depending on the mode.

        - 'sGaussian':
            sigma : float
                FWHM of the Gaussian function in wavenumber cm^(-1).
            k : float
                Controls the skewing of the asymmetry.
            a_sigma, a_k : float
                Tuning factors for sigma and k.
        - 'Voigt':
            sigma_V_l : float
                FWHM of the Voigt function in wavenumber cm^(-1) for
                the lower half.
            sigma_L_l : float
                FWHM of the Lorentzian function in wavenumber cm^(-1) for
                the lower half.
            sigma_V_h : float
                FWHM of the Voigt function in wavenumber cm^(-1) for
                the higher half.
            sigma_L_h : float
                FWHM of the Lorentzian function in wavenumber cm^(-1) for
                the higher half.
    offset : float
        Background offset.
    mode : 'sGaussian', str, optional
        Two options for the ISRF:

        - Asymmetric super Gaussian: 'sGaussian'.
        - Asymmetric Voigt: 'Voigt'.

    Returns
    -------
    1d array of floats
        Intensities of the peak-normalized asymmetric ISRF.
    """
    slit_fc = []
    if mode == 'sGaussian':
        slit_fc = asym_Gaussian(w, w0, param_1, param_2, param_3,
                                param_4, offset)
    elif mode == 'Voigt':
        slit_fc = asym_Voigt(w, w0, param_1, param_2, param_3, param_4,
                             param_5, param_6, offset)

    return slit_fc



"""Least-square fit of experimental CARS spectra."""
from pathlib import Path
from functools import wraps
import numpy as np
from .cars_synth import CarsSpectrum
from .convol_fcn import asym_Gaussian, asym_Voigt
from .utils import pkl_dump, pkl_load, downsample

try:
    from lmfit import Model
    from lmfit.printfuncs import report_fit
    _HAS_LMFIT = True
except Exception:
    _HAS_LMFIT = False


def _ensureLmfit(function):
    if _HAS_LMFIT:
        @wraps(function)
        def wrapper(*args, **kwargs):
            return function(*args, **kwargs)
        wrapper.__doc__ = function.__doc__
        return wrapper

    def no_op(*args, **kwargs):
        _message = ("lmfit module is required for carrying out the built-in "
                    "least-square fit. Please install lmfit first or build "
                    "custom fit routines based on cars_expt_sync().")
        raise Exception(_message)
    no_op.__doc__ = function.__doc__
    return no_op


def bg_removal(spec, bg=None):
    """Subtract background from spectrum and normalize it by its peak.

    Parameters
    ----------
    spec : 1d array
        Spectrum.
    bg : None, 1d array, optional
        Background noise.

    Return
    ------
    id array
        Background-subtracted, peak-normalized spectrum.
    """
    if bg is not None:
        _spec = (np.array(spec) - np.array(bg)).flatten()
    else:
        _spec = np.array(spec).flatten()

    return _spec/_spec.max()


@_ensureLmfit
def slit_fit(nu, spec, init_params=None, lineshape='sGaussian',
             eval_only=False, save_fit=False, dir_save=None, file_name=None):
    r"""fitting the experimental slit function with a chosen lineshape

    .. attention::
        It is recommended to always look for initial parameters by using the
        `eval_only` option to roughly match the shape of the experimental slit
        function. The built-in initial parameters may not work for all the
        cases.

    Parameters
    ----------
    nu : 1d array of floats
        Spectral positions in [:math:`\mathrm{cm}^{-1}`].
    spec : 1d array of floats
        Experimentally obtained slit function, usually an isolated atomic line
        from a calibration lamp.
    init_params : dict, optional
        Initial fitting parameters for the lineshape. Refer to
        :mod:`carspy.convol_fcn.asym_Voigt` and
        :mod:`carspy.convol_fcn.asym_Gaussian` for the required arguments.
    lineshape : str, optional
        Type of the lineshape, by default 'sGaussian'. Choose between
        'sGaussian' and 'sVoigt'.
    eval_only : bool, optional
        If true, returns the evaluation with given initial parameters.
    save_fit : bool, optional
        If true, fit results will be saved in a pickle file.
    dir_save : str, optional
        If `save_fit` is true, a string to the desired directory for saving.
    file_name : str, optional
        If `save_fit` is true, specify the file name without file extension.

    Returns
    -------
    1d array
        If `eval_only` is true, the evaluation result is returned.

    """
    if lineshape == 'sGaussian':
        fit_model = Model(asym_Gaussian, independent_vars='w')
    elif lineshape == 'sVoigt':
        fit_model = Model(asym_Voigt, independent_vars='w')
    else:
        raise ValueError("Please choose between 'sGaussian' and 'sVoigt'")

    if init_params is None:
        init_params = (
            ('w0', 0, True),
            ('sigma', 2, True, 0.01, 10),
            ('k', 2, True, 0, 10),
            ('a_sigma', -0.8, True, -5, 5),
            ('a_k', 1, True, -5, 5),
            ('sigma_L_l', 0.1, True, 0.01, 5),
            ('sigma_L_h', 0.1, True, 0.01, 5),
            ('offset', 0, True, 0, 1)
        )

    params = fit_model.make_params()
    params.add_many(*init_params)
    if eval_only:
        return fit_model.eval(params, nu=nu)
    else:
        fit_result = fit_model.fit(spec/spec.max(), params, w=nu)
        report_fit(fit_result, show_correl=True, modelpars=params)
        fit_result.plot()

        if save_fit:
            if dir_save and file_name:
                path_write = Path(dir_save) / (file_name + '.pkl')
                pkl_dump(path_write, fit_result)


class CarsFit():
    """Fitting experimental CARS spectrum.

    .. note::
        It can also be used to fit the laser linewidth or slit function.

    """

    def __init__(self, spec_cars, nu_spec, ref_fac=100., bg_cars=None,
                 spec_stokes=None, bg_stokes=None, fit_mode=None, **kwargs):
        r"""Input measured CARS spectrum and its background.

        Parameters
        ----------
        spec_cars : 1d array of floats
            Measured CARS spectrum.
        nu_spec : 1d array of floats
            Spectral range (axis) of the supplied CARS spectrum in
            [:math:`\mathrm{cm}^{-1}`]. This can either be absolute values or
            Raman shifted ones.
        ref_fac : float
            Refining factor based on the supplied spectral domain, by default
            100. This factor is used in the fitting procedure to synthesize
            spectra with much higher spectral resolution to be compared (after
            downsampling) with the supplied (expt.) CARS spectrum. Higher value
            will result in more accurate spectra at the cost of CPU time.
        bg_cars : 1d array of floats
            Background noise for the measured CARS spectrum.
        spec_stokes : 1d array of floats
            Measured broadband Stokes profile (usually with Ar), by default
            None. If not supplied, it is assumed that `spec_cars` is already
            corrected by the Stokes profile.
        bg_stokes : 1d array of floats
            Background noise for the measured Stokes profile.
        fit_mode : dict, optional
            A dictionary containing the control parameters used to perform the
            CARS fit. By default:

            power_factor : 0
                0 (fit I) or 1 (fit sqrt(I)).
            downsample : 'local_mean'
                Choose between 'local_mean' (highly efficient custom algorithm)
                and 'interp' (interpolation with :mod:`numpy.interp`).
            slit : 'sVoigt'
                Choose between (asymmetric) 'sVoigt' and 'sGaussian' as the
                slit impulse response function, see the documentations
                for :mod:`carspy.convol_fcn.asym_Voigt` and
                :mod:`carspy.convol_fcn.asym_Gaussian`. 'sGaussian' will be
                deprecated in future updates.
            pump_ls : 'Gaussian'
                'Gaussian' or 'Lorentzian' laser lineshape.
            chi_rs : 'G-matrix'
                Choose between 'isolated' and 'G-matrix' for the consideration
                of pressure effects.
            convol : 'Kataoka'
                'Kataoka'/'K' (double convolution) or 'Yuratich'/'Y' (single
                convolution).
            doppler_effect : True
                Whether or not to consider Doppler effect on the Raman
                lineshape.
            chem_eq : False
                Whether or not to assume chemical equilibrium during the fit.
            fit : 'custom'
                Type of build-in fitting setups: 'T_x' or 'custom'.

                - 'T_x': fitting variables related to the experimental setup
                  (e.g., spectrometer) are inheritted from an existing fit
                  and fixed. Only temperature and species concentrations are
                  by default allowed to vary.
                - 'custom': all fitting variables need to be provided before a
                  fit can process.
                See :mod:`carspy.cars_fit.CarsFit.ls_fit` for details.


        Other Parameters
        ----------------
        **kwargs:
            This method also allows the keyword arguments found for
            initializing :mod:`carspy.cars_synth.CarsSpectrum`.
        """
        # settings for the fit
        if fit_mode is None:
            self.fit_mode = {'power_factor': 0,
                             'downsample': 'local_mean',
                             'slit': 'Voigt',
                             'pump_ls': 'Gaussian',
                             'chi_rs': 'G-matrix',
                             'convol': 'Kataoka',
                             'doppler_effect': True,
                             'chem_eq': False,
                             'fit': 'custom'
                             }
        else:
            self.fit_mode = fit_mode

        # create subset of synth mode for CarsSpectrum
        self.synth_mode = {'pump_ls': self.fit_mode['pump_ls'],
                           'chi_rs': self.fit_mode['chi_rs'],
                           'convol': self.fit_mode['convol'],
                           'doppler_effect': self.fit_mode['doppler_effect'],
                           'chem_eq': self.fit_mode['chem_eq'],
                           }

        # subtract background (optional) and normalize by max
        self.spec_cars = bg_removal(spec_cars, bg_cars)
        if spec_stokes is not None:
            self.spec_stokes = bg_removal(spec_stokes, bg_stokes)
            if len(self.spec_stokes) != len(self.spec_cars):
                raise ValueError("The length of spec_cars needs to be "
                                 "identical to that of spec_stok")
        else:
            self.spec_stokes = spec_stokes

        # fixed properties
        self.nu = nu_spec
        self.ref_fac = ref_fac

        if len(self.nu) != len(self.spec_cars):
            raise ValueError("The length of spec_cars needs to be identical "
                             "to that of nu_spec")

        # setup CarsSpectrum
        self.spec_synth = CarsSpectrum(**kwargs)

        # define fit result
        self.fit_result = []

    def preprocess(self, w_Stokes=0, nu_Stokes=0, crop=None,
                   bg_subtract=False, bg_offset=0, bg_loc=None):
        r"""Prepare the raw data for the fitting.

        Parameters
        ----------
        w_Stokes : float, optional
            Center wavelength of the Stokes (e.g., dye laser) beam in [nm], by
            default 0.
        nu_shift : float, optional
            Center wavenumber of the Stokes beam in [:math:`\mathrm{cm}^{-1}`],
            by default 0.
            In essence, `w_Stokes` and `nu_Stokes` are equivalent.
        crop : list, optional
            Two indices to crop the spectrum with, by default None. Needs to be
            adjusted based on the experimental setup.
        bg_subtract : bool, optional
            If True, an extra offset specified by `bg_offset` or determined
            within `bg_loc` is subtracted from the spectrum.
            This is not recommended as there shouldn't be any physical
            background left if backgrounds are subtracted properly from the
            experimental spectrum beforehand. This might help if S/N is bad.
            By default it is set to False.
        bg_offset : float, optional
            Value used as background to subtract, be default 0. This is ignored
            if `bg_log` is provided.
        bg_loc : list, optional
            Two indices to select the part of spectrum as background, only used
            if `bg_subtract` is True.
        """
        # remove very small values in the argon spectrum
        if self.spec_stokes is not None:
            self.spec_stokes[self.spec_stokes <= 1e-3] = 1e-3
            self.spec_cars = self.spec_cars / self.spec_stokes

        # extra background removal (USE WITH CAUTION)
        if bg_subtract:
            if bg_loc is not None:
                _bg = self.spec_cars[bg_loc[0]:bg_loc[1]].mean()
            else:
                _bg = bg_offset
            self.spec_cars = self.spec_cars - _bg

        self.spec_cars[self.spec_cars < 0] = 0

        # crop signal and spectral axis and re-normalize
        if crop is not None:
            self.spec_cars = self.spec_cars[crop[0]:crop[1]]
            self.nu = self.nu[crop[0]:crop[1]]
        # take the square root if 'power_factor' is 1
        self.spec_cars = self.spec_cars**(0.5**self.fit_mode['power_factor'])
        self.spec_cars = self.spec_cars/self.spec_cars.max()

        # convert the spectral axis to relative wavenumber to match
        # the CARS program
        if w_Stokes != 0:
            self.nu = self.nu - 1e7/w_Stokes
        else:
            self.nu = self.nu - nu_Stokes

    def cars_expt_synth(self, nu_expt, x_mol, temperature, del_Tv, nu_shift,
                        nu_stretch, pump_lw,
                        param1, param2, param3, param4, param5, param6):
        r"""
        Synthesize a CARS spectrum based on the experimental spectral domain.

        Parameters
        ----------
        nu_expt : 1d array of floats
            The spectral axis determined in the experiment. This is used as
            the independent variable during the fit.
        x_mol : float
            Mole fraction of probed molecule.
        temperature : float
            Temperature in the probe volume in [K].
        nu_shift : float
            Shift applied to correctly center the spectrum in
            [:math:`\mathrm{cm}^{-1}`].
        nu_strech : float
            Strech applied to nu to compensate for incorrect dispersion
            calibration.
        pump_lw : float
            Pump laser linewdith in [:math:`\mathrm{cm}^{-1}`].
        del_Tv : float
            The amount vibrational temperature exceeds the rotational
            temperature.
        param1, param2, param3, param4, param5, param6 : float
            Fitting parameters for the slit function

        Return
        ------
        1d array
            Syntheiszed and downsampled CARS spectrum to match the length and
            spectral resolution of the measured spectrum. The spectrum is
            normalized by its peak value.
        """
        # shift the experimental spectral axis and refine it
        nu_expt = nu_expt*nu_stretch + nu_shift
        # using magnification to refine the grid
        _fine_factor = self.ref_fac
        _del_nu = nu_expt[1] - nu_expt[0]
        _nu_expt_pad = np.pad(nu_expt, (5, 5), 'reflect', reflect_type='odd')
        nu_f = np.arange(start=_nu_expt_pad[0], stop=_nu_expt_pad[-1],
                         step=_del_nu/_fine_factor)

        # calculate the slit function
        nu_slit = []
        if self.fit_mode['slit'] == 'sGaussian':
            nu_slit = asym_Gaussian(w=nu_f, w0=(nu_f[0]+nu_f[-1])/2,
                                    sigma=param1, k=param2,
                                    a_sigma=param3, a_k=param4, offset=0)
        elif self.fit_mode['slit'] == 'sVoigt':
            nu_slit = asym_Voigt(w=nu_f, w0=(nu_f[0]+nu_f[-1])/2,
                                 sigma=param1, k=param2,
                                 a_sigma=param3, a_k=param4,
                                 sigma_L_l=param5, sigma_L_h=param6,
                                 offset=0)
        # calculate the CARS spectrum
        _, I_as = self.spec_synth.signal_as(x_mol=x_mol,
                                            temperature=temperature, nu_s=nu_f,
                                            pump_lw=pump_lw, del_Tv=del_Tv,
                                            synth_mode=self.synth_mode)
        # convolute the slit function with the CARS spectrum
        I_as = np.convolve(I_as, nu_slit, 'same')

        # downsampling to the experimental spectral grid
        I_as_down = downsample(nu_expt, nu_f, I_as,
                               mode=self.fit_mode['downsample'])**(
                                   0.5**self.fit_mode['power_factor'])
        return np.nan_to_num(I_as_down/I_as_down.max())

    @_ensureLmfit
    def ls_fit(self, add_params=None, path_fit=None, show_fit=False,
               eval_only=False, **kwargs):
        """
        Fitting the experimental CARS spectrum.

        .. attention::
            The least-quare fit module ``lmfit`` is necessary for this method.
            Please be aware that certain Python versions may not be supported
            by ``lmfit``. For displaying the fit results ``matplotlib`` will be
            needed as well.

        Parameters
        ----------
        add_params : nested tuple, optional
            List of parameters controlling the fitting process. This option can
            be used to modify these initial parameters:

            .. code-block:: python

                (('temperature', 2000, True, 250, 3000),
                 ('del_Tv', 0, False),
                 ('x_mol', 0.6, x_mol_var, 0.2, 1.5),
                 ('nu_shift', fit_params['nu_shift'], False),
                 ('nu_stretch', fit_params['nu_stretch'], False),
                 ('pump_lw', fit_params['pump_lw'], False),
                 ('param1', fit_params['param1'], False),
                 ('param2', fit_params['param2'], False),
                 ('param3', fit_params['param3'], False),
                 ('param4', fit_params['param4'], False),
                 ('param5', fit_params['param5'], False),
                 ('param6', fit_params['param6'], False))
            Each element of the nested tuple has the following element in
            order:
                variable_name : str
                    All the arguments of
                    :mod:`carspy.cars_fit.CarsFit.cars_expt_synth`
                    are admissible variables except for the independent
                    variable `nu_expt`.
                initial_guess : float
                    Initial guess or fixed value set for this variable.
                variable : bool
                    Determine if the variable is fixed (False) or not (True)
                    during the fit.
                lower_bound : float
                    Lower boundary for the fitting variable. If not provided,
                    negative infinity will be assumed.
                upper_bound : float
                    Upper boundary for the fitting variable. If not provide,
                    positive infinity will be assumed.
            For more details refer to the documentation of ``lmfit.Model``.
        path_fit : str
            Path to the `.pkl` file of fitting result created by
            :mod:`carspy.cars_fit.CarsFit.save_fit`. This allows importing
            the fitting result of an existing spectrum, such that the inferred
            values of certain parameters (such as those related to the
            spectrometer) could be re-used in the next fit. A standard use case
            for this would be the fitting result of a room-temperature
            spectrum. This is needed if the `fit` in `fit_mode` of
            :mod:`carspy.cars_fit.CarsFit` is set to `T_x`.
        show_fit : bool, optional
            If True, the fitting results will be reported and plotted. This is
            done via built-in functions in ``lmfit``.
        """
        # general setup
        fit_model = Model(self.cars_expt_synth, independent_vars=['nu_expt'])
        initi_params = []
        # different fitting modes
        if self.fit_mode['fit'] == 'T_x':
            if path_fit is None:
                raise ValueError("Please provide path to a .pkl file "
                                 "containing the fitting result of a spectrum")
            if self.fit_mode['chem_eq']:
                x_mol_var = False
            else:
                x_mol_var = True
            fit_params = pkl_load(path_fit).params
            initi_params = (('temperature', 2000, True, 250, 3000),
                            ('del_Tv', 0, False),
                            ('x_mol', 0.6, x_mol_var, 0.2, 1.5),
                            ('nu_shift', fit_params['nu_shift'], False),
                            ('nu_stretch', fit_params['nu_stretch'], False),
                            ('pump_lw', fit_params['pump_lw'], False),
                            ('param1', fit_params['param1'], False),
                            ('param2', fit_params['param2'], False),
                            ('param3', fit_params['param3'], False),
                            ('param4', fit_params['param4'], False),
                            ('param5', fit_params['param5'], False),
                            ('param6', fit_params['param6'], False))

        if self.fit_mode['fit'] == 'custom':
            if add_params is None:
                raise ValueError("Please specify fitting parameters first "
                                 "using add_params")
        params = fit_model.make_params()
        params.add_many(*initi_params)
        if add_params is not None:
            params.add_many(*add_params)

        if eval_only:
            return fit_model.eval(params, nu_expt=self.nu)
        else:
            self.fit_result = fit_model.fit(np.nan_to_num(self.spec_cars),
                                            params,
                                            nu_expt=self.nu, **kwargs)

            if show_fit:
                report_fit(self.fit_result, show_correl=True, modelpars=params)
                self.fit_result.plot()

    def save_fit(self, dir_save, file_name):
        """
        Save the fitting results in a pickle file.

        Parameters
        ----------
        dir_save : path
            A valid local directory.
        file_name : str
            File name of the pickle file to be saved.
        """
        path_write = Path(dir_save) / (file_name + '.pkl')
        pkl_dump(path_write, self.fit_result)
